#include <wx/progdlg.h>
#include "dfarc.hpp"
#include "stdio.h"

//
// wxString Bzip::Create(wxGauge *PBar = 0)
//
// Compresses a Bzip file.  Takes filename to compress from mFilePath.
//
//     PBar - Parameter passed to a pre-made wxGauge displaying the progress of the decompression.
//
// Returns: Compressed file location.
//
wxString Bzip::Create(wxProgressDialog* aProgressDialog)
{
    // First, grab a random filename.
    wxString strTempFile = GetTempName();
    
    // Open temp file.
    FILE* fpOut = fopen(strTempFile.c_str(), "wb");
    FILE* fpIn;
    BZFILE* bzOut = 0;
    int iErr = 0, iNBytes = 0, iTotalBytes = 0, iBytesWritten = 0;

    if(!fpOut)
    {
        ::wxLogError("Error: Could not open output file '%s' for bzip compression.", strTempFile.c_str());
        return "ERROR";
    }

    bzOut = BZ2_bzWriteOpen(&iErr, fpOut, 9, 0, 90);

    if(iErr != BZ_OK)
    {
        ::wxLogError("Error: Could not initialize compression method!  Will not generate a correct .dmod file.  Quitting.");
        fclose(fpOut);
        return "ERROR";
    }


    // Open up the input file
    fpIn = fopen(mFilePath.c_str(), "rb");

    if(!fpIn)
    {
        ::wxLogError("Error: Could not open input file '%s' for bzip compression.", mFilePath.c_str());
        fclose(fpOut);
        return "ERROR";
    }
    
    // Get the file size. (I hate C I/O)
    fseek(fpIn, 0, SEEK_END);
    int iEnd = ftell(fpIn);
    fseek(fpIn, 0, SEEK_SET);
    int iBeg = ftell(fpIn);
    iTotalBytes = iEnd - iBeg;

    unsigned char Buf[4096];

    // Keep getting chunks of the file
    while(1)
    {
        memset(Buf, 0, 4096);

        iNBytes = fread(Buf, 1, 4096, fpIn);
        iBytesWritten += iNBytes;

        BZ2_bzWrite(&iErr, bzOut, Buf, iNBytes);

        // Done
        if(feof(fpIn))
        {
            break;
        }
        
        
        if ( aProgressDialog != NULL )
        {
            int lPercent( ( 100 * iBytesWritten ) / iTotalBytes );
            aProgressDialog->Update( lPercent, "" );

#ifndef WIN32
	    // Windows has no problems yielding events.  Linux does.
	    // Only show every 10% or so, since it's slightly intensive.
	    if ( lPercent % 10 == 0 )
	    {
	        ::wxYield();
	    }
#endif
        }
    }

    // Close up.
    BZ2_bzWriteClose(&iErr, bzOut, 0, 0, 0);

    fclose(fpIn);
    fclose(fpOut);

    return strTempFile;
}

//
// wxString Bzip::Extract(wxGauge *PBar = 0)
//
// Decompresses a Bzip file.  Takes filename to decompress from mFilePath.
//
//     PBar - Parameter passed to a pre-made wxGauge displaying the progress of the decompression.
//
// Returns: Decompressed file location.
//
wxString Bzip::Extract( wxProgressDialog* aProgressDialog )
{
    // First, grab a random filename.
    wxString lTempFileName = GetTempName();
    int lError( 0 );
    unsigned long lTotalBytes( 0 );
    unsigned long lTotalCurrentBytesWrote( 0 );

    // Begin decompress stuff.
	FILE* lDmodFileIn = fopen( mFilePath.c_str(), "rb" );
    
    if ( lDmodFileIn == NULL )
    {
        ::wxLogError("Error: Could not open input file '%s' for bzip decompression.", mFilePath.c_str());
        return "";
    }

    // Get the file size
    fseek( lDmodFileIn, 0, SEEK_END );
    unsigned long lEnd = ftell( lDmodFileIn );
    fseek( lDmodFileIn, 0, SEEK_SET );
    lTotalBytes = lEnd - ftell( lDmodFileIn );

    BZFILE* lBzipFileIn = BZ2_bzReadOpen( &lError, lDmodFileIn, 0, 0, 0, 0 );

    if( lError != BZ_OK )
    {
        ::wxLogError( "Error: Invalid .dmod file selected!" );
		fclose( lDmodFileIn );
        return "";
    }

    // Open up the output file
    FILE* lTarFileOut = fopen( lTempFileName.c_str(), "wb" );
    
    if( lTarFileOut == NULL )
    {
        ::wxLogError("Error: Could not write to '%s'.", lTempFileName.c_str());
        fclose( lDmodFileIn );
        return "";
    }

    // Keep getting chunks of the file
    while( true )
    {
        unsigned char lBuffer[4096];
        memset( lBuffer, 0, 4096 );
        int lCurrentBytesRead( BZ2_bzRead( &lError, lBzipFileIn, lBuffer, 4096 ) );
        fwrite( lBuffer, lCurrentBytesRead, 1, lTarFileOut );
        lTotalCurrentBytesWrote += lCurrentBytesRead;

        if ( lError != BZ_OK )
        {
            // Success
            if ( lError == BZ_STREAM_END )
            {
                break;
            }
            else if ( lError == BZ_PARAM_ERROR )
            {
                ::wxLogError( "Invalid .dmod file specified." );
            }
            else if ( lError == BZ_SEQUENCE_ERROR )
            {
                ::wxLogError( "Critical program function error: opened for write." );
            }
            else if ( lError == BZ_IO_ERROR )
            {
                ::wxLogError( "Could not read .dmod file." );
            }
            else if ( lError == BZ_UNEXPECTED_EOF )
            {
                ::wxLogError( "Incomplete .dmod file. Please download it again." );
            }
            else if ( lError == BZ_DATA_ERROR )
            {
                ::wxLogError( "The .dmod file is corrupted.  Please download it again." );
            }
            else if ( lError == BZ_DATA_ERROR_MAGIC )
            {
                ::wxLogError( "The file is not a valid .dmod file." );
            }
            else if ( lError == BZ_MEM_ERROR )
            {
                ::wxLogError( "Out of memory error." );
            }
            else
            {
                ::wxLogError( "An unhandled error occured." );
            }
            remove( lTempFileName );
            lTempFileName = "";
            break;
        }
        if ( aProgressDialog != NULL )
        {
            int lPercent = 16 * lTotalCurrentBytesWrote / lTotalBytes;
            if ( lPercent >= 100 )
            {
                lPercent = 99;
            }
            aProgressDialog->Update( lPercent );
        }
        ::wxYield();
    }
    aProgressDialog->Update( 100 );
    BZ2_bzReadClose( &lError, lBzipFileIn );
    fclose( lDmodFileIn );
    fclose( lTarFileOut );

    return lTempFileName;
}
