#include <wx/progdlg.h>
#include "dfarc.hpp"

//
// TAR CREATE FUNCTIONS
//
wxString Tar::Create( wxProgressDialog* aProgressDialog, const wxString& aFolderName, const bool& aIncludeSaveDat, const bool& aIncludeThumbsDb )
{         
    if(!bCanCompress)
    {
        wxString str;
        return str;
    }    
        
    wxArrayString wxasFileList;
    int iNumEntries = 0;
    float fPMVal = 0;
    

    // Our fstream settings are set to append to the end of the tar file.
    // This could screw it up if it is run again.  Delete it if it exists.
    if( ::wxFileExists( mFilePath ) )
    {
        remove( mFilePath.c_str() );
    }

    // Populate file list.
    if(DFSPtr == 0)
    {
        ::wxLogFatalError("Error: DFilesystem* == 0 error in Tar::Create().");
        throw;
    }
    else
    {
        DFSPtr->GetAllFiles(wxasFileList, strCompressDir);
    }
    
    // We'll make the paths relative now
    for(register int i = 0; i < wxasFileList.GetCount(); ++i)
    {
        wxFileName *wxfn = new wxFileName(wxasFileList.Item(i));
        wxfn->MakeRelativeTo(strCompressDir);
        wxasFileList.Item(i) = wxfn->GetFullPath(wxPATH_UNIX);
        delete wxfn;
    }

    // Save this for later (Progress Bar)
    iNumEntries = wxasFileList.GetCount();
    
    
    bool bFirstEntry = true;
    TarHeader th;
    std::fstream f_In, f_Out;
    
    
    // 8KB buffer
    char *szBuf = new char[8192];
    
        
    // Open output file
    f_Out.open(mFilePath.c_str(), std::ios::out | std::ios::binary);

    if(!f_Out.good())
    {
        ::wxLogFatalError("Error: File '%s' not found!", mFilePath.c_str());
        throw;
    }

    for(register int i = 0; i < wxasFileList.GetCount(); ++i)
    {
        // Write tar data.
        memset(&th, 0, sizeof(TarHeader));
        memset(szBuf, 0, sizeof(szBuf));
        
        FillHeader( wxasFileList.Item(i), aFolderName, th );
        int iAmountToRound = 0;
        
        // Find the absolute path of the input file.
        wxString str = strCompressDir + wxasFileList.Item(i);
        
        if(!bFirstEntry)
        {
			// Round to the nearest 512-byte mark.
            iAmountToRound = RoundTo512(f_Out.tellp()) - f_Out.tellp();
            
            for(int i = 0; i < iAmountToRound; ++i)
            {
                f_Out.put('\0');
            }
        }
        
        // Make sure we know it's not our first entry.
        bFirstEntry = false;

        // Now, write our record.
        f_Out.write(reinterpret_cast<char*>(&th), sizeof(TarHeader));

        // Open our input file.
        f_In.open(str.c_str(), std::ios::in | std::ios::binary);
    
        if(!f_In.good())
        {
            ::wxLogFatalError("Error: File '%s' not found!  Cannot archive file.", str.c_str());
            throw;
        }
        
        // Write the file.
        while(f_In.peek() != EOF)
        {
            memset(szBuf, 0, sizeof(szBuf));    
            
            f_In.read(szBuf, sizeof(szBuf));
            f_Out.write(szBuf, sizeof(szBuf));
        }

        
        // Update the progress bar.
        if ( aProgressDialog != 0 )
        {
            // Update the meter.
            int lPercent( ( 100 * i ) / iNumEntries );
            aProgressDialog->Update( lPercent, "" );

#ifndef WIN32
	    // Windows has no problems yielding events.  Linux does.
	    // Only show every 10% or so, since it's slightly intensive.
	    if ( lPercent % 10 == 0 )
	    {
	        ::wxYield();
	    }
#endif
        }
        
        f_In.close();
        f_In.clear();
    }
        
    // BE FREEEEEEE!!!
    delete[] szBuf;
	
	
	// Write the end header.
	int iAmountToRound = RoundTo512(f_Out.tellp()) - f_Out.tellp() + 512;

	for(int i = 0; i < iAmountToRound; ++i)
	{
		f_Out.put('\0');
	}
    
    
    // Close the output file.
    f_Out.close();
    
        
    return mFilePath;
}



void Tar::FillHeader(wxString &mFilePath, const wxString& aDestinationFile, TarHeader &Record)
{
    int iFileBegin =  0;
    int iFileEnd = 0;
    int iChksum = 0;
    bool bIsDir = true;
    char *RecordPtr;
    std::fstream f_In;


    // Find the absolute path.
    wxString str = strCompressDir + mFilePath;

    if(!::wxDirExists(str))
    {
	bIsDir = false;

        // Open the file.
        f_In.open(str.c_str(), std::ios::in | std::ios::binary);

        if(!f_In.good())
        {
            ::wxLogError("Error: File '%s' not found!  Cannot get size.", str.c_str());
            throw;
        }

        // Get the file size.
        iFileBegin = f_In.tellg();
        f_In.seekg(0, std::ios::end);
        iFileEnd = f_In.tellg();
        f_In.seekg(0, std::ios::beg);
        f_In.close();
        f_In.clear();
    }

    // Start collecting data.
    strncpy(Record.Name, aDestinationFile.Lower() + "/" + mFilePath.Lower().c_str(), 100);

    #ifdef WIN32
        // No UID or GID support for Windows - Hack it up.
        strncpy(Record.Mode, "0100666", 8);
        strncpy(Record.Uid, "0000000", 8);
        strncpy(Record.Gid, "0000000", 8);
    #else
        int iMode = 0;
        struct stat sFileInfo;

        stat(str.c_str(), &sFileInfo);

        if(!(sFileInfo.st_mode & S_IFREG))
        {
            // Not a normal file.  Let's archive it anyway, but warn the user.
            ::wxLogWarning("Warning: Archiving non-ordinary file (Linked or Device file?).");
        }

        // Add the normal file bit.
        iMode += 100000;

        //  Find out the file pernissions.
        if(sFileInfo.st_mode & 00400)
        {
            iMode += 400;
        }
        if(sFileInfo.st_mode & 00200)
        {
            iMode += 200;
        }
        if(sFileInfo.st_mode & 00100)
        {
            iMode += 100;
        }
        if(sFileInfo.st_mode & 00040)
        {
            iMode += 40;
        }
        if(sFileInfo.st_mode & 00020)
        {
            iMode += 20;
        }
        if(sFileInfo.st_mode & 00010)
        {
            iMode += 10;
        }
        if(sFileInfo.st_mode & 00004)
        {
            iMode += 4;
        }
        if(sFileInfo.st_mode & 00002)
        {
            iMode += 2;
        }
        if(sFileInfo.st_mode & 00001)
        {
            iMode += 1;
        }


        {
            char *sz = new char[8];
            
            // Write it to the Mode header.
            strncpy(sz, ToString(iMode), 8);
            
            if(!PadWithZeros(sz, 8, 8))
            {
		// This should like...never happen.  Logic error.
		::wxLogFatalError("Error: Logic error in Tar::PadWithZeros.  Call made from Tar::FillHeader.");
                delete[] sz;
                throw;
            }
            else
            {
                strncpy(Record.Mode, sz, 8);
            }
    
            // Write the UID and GID.
            strncpy(sz, ToString(sFileInfo.st_uid), 8);
            
            if(!PadWithZeros(sz, 8, 8))
            {
		// This should like...never happen.  Logic error.
		::wxLogFatalError("Error: Logic error in Tar::PadWithZeros.  Call made from Tar::FillHeader.");
                delete[] sz;
                throw;
            }
            else
            {
                strncpy(Record.Uid, sz, 8);
            }
    
            strncpy(sz, ToString(sFileInfo.st_gid), 8);
            
            if(!PadWithZeros(sz, 8, 8))
            {
		// This should like...never happen.  Logic error.
		::wxLogFatalError("Error: Logic error in Tar::PadWithZeros.  Call made from Tar::FillHeader.");
                delete[] sz;
                throw;
            }
            else
            {
                strncpy(Record.Gid, sz, 8);
            }
            
            delete[] sz;
        }
    #endif

    // If we're a directory, we don't need to go any farther.
    if(bIsDir)
    {
        return;
    }

    // File Size.
    char *sz = new char[12];
    char *szPtr = sz;
    DecToOct(iFileEnd - iFileBegin, sz, 12);
    
    if(!PadWithZeros(sz, 12, 12))
    {
	// This should like...never happen.  Logic error.
	::wxLogFatalError("Error: Logic error in Tar::PadWithZeros.  Call made from Tar::FillHeader.");
        delete[] sz;
        throw;
    }
    else
    {
        strncpy(Record.Size, sz, 12);
    }    
    
    // Time of modification
    memset(sz, 0, 12);
    DecToOct(time(0), sz, 12);
    
    if(!PadWithZeros(sz, 12, 12))
    {
	// This should like...never happen.  Logic error.
	::wxLogFatalError("Error: Logic error in Tar::PadWithZeros.  Call made from Tar::FillHeader.");
        delete[] sz;
        throw;
    }
    else
    {
        strncpy(Record.Mtime, sz, 12);
    }
    
    delete[] sz;

    // Add blank checksum.
    memset(Record.Chksum, 0, 8);

    // Type of file.
    Record.Linkflag = 0;
    strncpy(Record.Linkname, "", 100);
    strncpy(Record.Magic, "ustar  ", 8);

    // Windows again, let's cheat.
    #ifdef WIN32
        strncpy(Record.Uname, "root", 32);
        strncpy(Record.Gname, "root", 32);
    #else
        // Let's get the username and such, then.
        passwd *pwdFileInfo = getpwuid(sFileInfo.st_uid);
        group *grpFileInfo = getgrgid(sFileInfo.st_gid);

        if(!pwdFileInfo || !grpFileInfo)
        {
            // No entry.  Let's fake it.
            strncpy(Record.Uname, Record.Uid, 32);
            strncpy(Record.Gname, Record.Gid, 32);
        }
        else
        {
            strncpy(Record.Uname, pwdFileInfo->pw_name, 32);
            strncpy(Record.Gname, grpFileInfo->gr_name, 32);
        }
    #endif

    // Skip devminor and devmajor.
    strncpy(Record.Devmajor, "", 8);
    strncpy(Record.Devminor, "", 8);

    // Fill the padding.
    strncpy(Record.Padding, "", 167);

    // Initialize our pointer.
    RecordPtr = reinterpret_cast<char*>(&Record);

    // Find the checksum.
    for(RecordPtr; (RecordPtr - reinterpret_cast<char*>(&Record)) < 512; ++RecordPtr)
    {
        iChksum += static_cast<int>(*RecordPtr);
    }

    // Add 256 (2^8)
    iChksum += 256;

    // Get it to Octal.
    char *chksum = new char[8];
    char *chksumPtr = chksum;
    DecToOct(iChksum, chksum, 8);
    
    if(!PadWithZeros(chksum, 8, 8))
    {
	// This should like...never happen.  Logic error.
	::wxLogFatalError("Error: Logic error in Tar::PadWithZeros.  Call made from Tar::FillHeader.");
        delete[] sz;
        throw;
    }
    else
    {
        strncpy(Record.Chksum, chksum, 8);
    }
    
    delete[] chksum;
    
    f_In.close();
}

int Tar::ReadHeaders( void )
{
    std::fstream f_In;
    TarHeader lHeader;
    TarRecord lRecord;
    unsigned int iBegData = 0;

    f_In.open(mFilePath.c_str(), std::ios::in | std::ios::binary | std::ios::beg);

    if(!f_In.good())
    {
        wxLogFatalError("Error: File '%s' not found!  Cannot read data.", mFilePath.c_str());
        return 1;
    }
	
	f_In.seekg(std::ios::beg);

    wxString lDmodDizPath;
    mmDmodDescription = "";
    mInstalledDmodDirectory = "";

    int lPosition( 0 );
    while(1)
    {
        memset(&lHeader, 0, sizeof(TarHeader));
        memset(&lRecord, 0, sizeof(TarRecord));

        // Read the data.
        f_In.read(reinterpret_cast<char*>(&lHeader), sizeof(TarHeader));

        iBegData += sizeof(TarHeader);

        if(!VerifyChecksum(&lHeader))
        {
            // Nope.  Exit.
            wxLogFatalError("Error: This .dmod file has an invalid checksum!  Cannot read file.");
            return 1;
        }
        
        strncpy(lRecord.Name, lHeader.Name, 100);

        // Bug fix for phantom files that aren't really files...
        if ( strcmp( lHeader.Name, "" ) == 0 )
        {
            lPosition += 512;
            continue;
        }

		// Size
		int lLength = sizeof( lHeader.Size );
		for ( int i = 1; i < lLength; i++ )
		{
			lRecord.iFileSize += ( lHeader.Size[ lLength - i - 1 ] - 0x30 ) * pow(8.0, static_cast<double>(i) - 1.0);
		}
        lRecord.iFilePosBegin = iBegData;
        
        // Some tar files have a multitude of NULL characters at the end to round up to the 10K mark.
        // If the header is blank, we don't have any more data.
        if(strcmp(lHeader.Name, "") == 0)
        {
            break;
        }
		//wxLogError("%s %d", lHeader.Name, lRecord.iFileSize);

        wxString lPath( lRecord.Name );
        if ( mInstalledDmodDirectory.Length() == 0 )
        {
            mInstalledDmodDirectory = lPath.SubString( 0, lPath.Find( '/' ) ).c_str();
            lDmodDizPath = mInstalledDmodDirectory + "dmod.diz";
            lDmodDizPath.LowerCase();
        }
        if ( lPath.Lower() == lDmodDizPath )
        {
            f_In.seekg( lRecord.iFilePosBegin, std::ios::beg );
            for ( int j = 0; j < lRecord.iFileSize; ++j )
            {
                mmDmodDescription += f_In.get();
            }
        }

        TarRecords.push_back(lRecord);

        // bug fix... files with a size of 0 still have a 512 chunk allocated for them
        //if ( lRecord.iFileSize == 0 )
        //{
        //    lRecord.iFileSize = 1;
        //}
        
        // Add the size to iBegData.
        iBegData += lRecord.iFileSize;
        
        // Move BegData to 512.
        iBegData = RoundTo512(iBegData);

        // Move forward to the next header.
        lPosition = RoundTo512(lRecord.iFileSize + lRecord.iFilePosBegin);
        f_In.seekg(lPosition, std::ios::beg);
    }

    f_In.close();
    
    return 0;
}

int Tar::Extract( wxProgressDialog* aProgressDialog )
{
    wxString strBuf;
    wxString strCwd = ::wxGetCwd();
    std::fstream f_In;
    int lError( 0 );

    // Prepare to chdir.
    DFSPtr->FixSlashes(strCwd);
    
    // Open the file here so it doesn't error after changing.
    f_In.open(mFilePath.c_str(), std::ios::in | std::ios::binary);
    
    if(!f_In.good())
    {
        lError = 1;
        ::wxLogFatalError("Error: File '%s' not found!  Cannot extract data.", mFilePath);
        throw;
    }
    
    // Move to the directory.
	chdir(DFSPtr->getDinkPath());

    // Put the data in the directories.
    if ( ExtractData( f_In, aProgressDialog ) != 0 )
    {
        lError = 1;
    }
	f_In.close();
	f_In.clear();

    // We're done.  Move back.
    chdir(strCwd.c_str());
    
    return lError;
}

int Tar::ExtractData( std::fstream& aTarStreamIn, wxProgressDialog* aProgressDialog )
{
    std::fstream lStreamOut;
    int lError( 0 );

    unsigned long lTotalBytes( 0 );
    unsigned long lTotalBytesRead( 0 );

    // Get the file size
    aTarStreamIn.seekg( 0, std::ios::end );
    unsigned long lEnd = static_cast<unsigned long>( aTarStreamIn.tellg() );
    aTarStreamIn.seekg( 0, std::ios::beg );
    lTotalBytes = lEnd - static_cast<unsigned long>( aTarStreamIn.tellg() );

    // Move into the extract dir.
	std::string lPreviousWorkingDirectory( ::wxGetCwd() );
	chdir( DFSPtr->getDinkPath() );

    // Extract the files.
    for ( unsigned int lTarRecordIndex = 0; lTarRecordIndex < TarRecords.size(); ++lTarRecordIndex )
    {
		TarRecord lCurrentTarRecord = TarRecords.at( lTarRecordIndex );
		std::string lCurrentFilePath( lCurrentTarRecord.Name );
		std::string lCurrentDirectory( lCurrentFilePath.substr( 0, lCurrentFilePath.find_last_of( '/' ) ) );

        // Odd bad file problem...
        if ( lCurrentFilePath.compare( "" ) == 0 )
        {
            // Only warn them if it doesn't come at the end.
            if ( lTarRecordIndex != TarRecords.size() - 1 )
            {
                lError = 1;
                wxLogError( "Got bad file %d/%d.  Skipping.", lTarRecordIndex, TarRecords.size() );
            }
        }
        else
        {
		    // Create the directory if it doesn't exist
		    if ( ::wxDirExists( lCurrentDirectory.c_str() ) == false )
		    {
			    wxFileName::Mkdir( lCurrentDirectory.c_str(), 755, wxPATH_MKDIR_FULL );
		    }

            // Only write the file if it is a file... some files are stored as directories
            char lLastCharacter( lCurrentFilePath[lCurrentFilePath.length() - 1] );
            if ( lLastCharacter != '\\' && lLastCharacter != '/' )
            {
                // Open the file.
                lStreamOut.open( lCurrentFilePath.c_str(), std::ios::out | std::ios::binary );
                if( lStreamOut.good() == false )
                {
                    lStreamOut.close();
                    lStreamOut.clear();
			        wxLogError( "Error: Improperly archived file '%s'.  Skipping.", lCurrentFilePath.c_str() );
                    lError = 1;
                    continue;
                }
                
                // Write it (one byte at a time?  Seems slow...)
		        aTarStreamIn.seekg( lCurrentTarRecord.iFilePosBegin, std::ios::beg );
                for ( int j = 0; j < lCurrentTarRecord.iFileSize; ++j )
                {
                    lStreamOut.put( aTarStreamIn.get() );
                }
                lTotalBytesRead += lCurrentTarRecord.iFileSize;
                
                // Close up.
                lStreamOut.close();
                lStreamOut.clear();
            }
            
            // Update progress
            if( aProgressDialog != NULL )
            {
                double lPercent( 100.0 * lTotalBytesRead / lTotalBytes ) ;
                if ( lPercent >= 100 )
                {
                    lPercent = 99.0;
                }
                aProgressDialog->Update( static_cast<int>( lPercent ), lCurrentFilePath.c_str() );
                ::wxYield();
            }
        }
    }
    aProgressDialog->Update( 100, "Done." );
	chdir( lPreviousWorkingDirectory.c_str() );
    return lError;
}

wxString Tar::getInstalledDmodDirectory( void )
{
    return mInstalledDmodDirectory;
}

wxString Tar::getmDmodDescription( void )
{
    return mmDmodDescription;
}